﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Primitives3D;
using XNA4;

namespace XNA4Primitives3DEffect
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont spriteFont;

        SampleArcBallCamera camera;
        SampleGrid grid;

        struct SaveStateMode
        {
            BlendState m_Blenstate;
            DepthStencilState m_DepthState;
            RasterizerState m_RasterState;
            SamplerState m_SamplerState;

            GraphicsDevice m_Device;

            public SaveStateMode(GraphicsDevice device)
            {
                m_Device = device;

                m_Blenstate = m_Device.BlendState;
                m_DepthState = m_Device.DepthStencilState;
                m_RasterState = m_Device.RasterizerState;
                m_SamplerState = m_Device.SamplerStates[0];
            }

            public void Restore()
            {
                m_Device.BlendState = m_Blenstate;
                m_Device.DepthStencilState = m_DepthState;
                m_Device.RasterizerState = m_RasterState;
                m_Device.SamplerStates[0] = m_SamplerState;
            }
        }

        /// <summary>
        /// input
        /// </summary>
        static InputHandler inputs;
        /// <summary>
        /// FPS Counter
        /// </summary>
        static FpsCounter fpsCounter;

        // effect
        Matrix projection;
        Matrix view;
        Matrix world;

        // Store a list of primitive models, plus which one is currently selected.
        List<GeometricPrimitive> primitives = new List<GeometricPrimitive>();
        int activePrimitive = 0;

        // rotate & zoom
        float angle = 0.0f;
        float zoom = 1.0f;

        // store a wireframe rasterize state
        RasterizerState wireFrameState;

        // Are we rendering in wireframe mode?
        bool isWireframe = true;
        bool isAnimation = true;

        // effect
        BasicEffect effect;
        Vector3 eyePosition = new Vector3(0.0f, 2.0f, 4.0f);  // same as LookAt

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here
            /// <summary>
            /// inputs component
            /// </summary>
            inputs = new InputHandler(this);
            Components.Add(inputs);
            /// <summary>
            /// fps component
            /// </summary>
            fpsCounter = new FpsCounter(this);
            Components.Add(fpsCounter);


            // Set up the reference grid
            grid = new SampleGrid();
            grid.GridColor = Color.LimeGreen;
            grid.GridScale = 1.0f;
            grid.GridSize = 32;
            // Set the grid to draw on the x/z plane around the origin
            grid.WorldMatrix = Matrix.Identity;
            // set up the sample camera
            camera = new SampleArcBallCamera(SampleArcBallCameraMode.RollConstrained);
            camera.Distance = 3;
            // orbit the camera so we're looking down the z=-1 axis, at the "front" of the object
            camera.OrbitRight(MathHelper.Pi);
            // orbit up a bit for perspective
            camera.OrbitUp(.2f);

            InitializeTransform();
            InitializeEffect();

            base.Initialize();
        }

        /// <summary>
        /// Initializes the transforms used for the scene
        /// </summary>
        private void InitializeTransform()
        {
            float tilt = (float)Math.PI / 8.0f;
            // Use the world matrix to tilt the cube along x and y axes.
            world = Matrix.CreateRotationX(tilt) * Matrix.CreateRotationY(tilt);
            view = Matrix.CreateLookAt(eyePosition, Vector3.Zero, Vector3.Up);
            projection = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.PiOver4,  // so this is 45 degrees.
                (float)GraphicsDevice.Viewport.Width /
                (float)GraphicsDevice.Viewport.Height,
                1.0f, 100.0f);

            grid.ProjectionMatrix = projection;
            grid.ViewMatrix = view;
        }

        /// <summary>
        /// Initializes the effect
        /// </summary>
        private void InitializeEffect()
        {
            // Initialize BasicEffect
            effect = new BasicEffect(graphics.GraphicsDevice);
            //effect.EnableDefaultLighting();
            effect.PreferPerPixelLighting = true;
            effect.Alpha = 1.0f;

            effect.DiffuseColor = Color.CornflowerBlue.ToVector3(); //  new Vector3(1.0f, 1.0f, 1.0f);    // object material
            effect.SpecularColor = Color.Yellow.ToVector3(); // new Vector3(1.0f, 1.0f, 0.0f);
            effect.SpecularPower = 16.0f;
            effect.AmbientLightColor = Color.DarkSlateGray.ToVector3(); // new Vector3(0.75f, 0.75f, 0.75f);

            effect.DirectionalLight0.Enabled = true;
            effect.DirectionalLight0.DiffuseColor = Vector3.One;
            effect.DirectionalLight0.Direction = Vector3.Normalize(new Vector3(1.0f, 1.0f, 1.0f));
            effect.DirectionalLight0.SpecularColor = Vector3.One;

            effect.DirectionalLight1.Enabled = true;
            effect.DirectionalLight1.DiffuseColor = Vector3.One;
            effect.DirectionalLight1.Direction = Vector3.Normalize(new Vector3(-1.0f, -1.0f, 1.0f));
            effect.DirectionalLight1.SpecularColor = Vector3.One;

            effect.LightingEnabled = true;
            
            effect.World = world;
            effect.View = view;
            effect.Projection = projection;
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>("hudfont");

            // TODO: use this.Content to load your game content here
            primitives.Add(new CubePrimitive(GraphicsDevice));
            primitives.Add(new SpherePrimitive(GraphicsDevice));
            primitives.Add(new CylinderPrimitive(GraphicsDevice));
            primitives.Add(new TorusPrimitive(GraphicsDevice));
            primitives.Add(new PyramidPrimitive(GraphicsDevice));

            wireFrameState = new RasterizerState()
            {
                FillMode = FillMode.WireFrame,
                CullMode = CullMode.None,
            };

            // Set up the reference grid and sample camera
            grid.LoadGraphicsContent(graphics.GraphicsDevice);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // TODO: Add your update logic here
            UpdateInput();

            GamePadState gpState = GamePad.GetState(PlayerIndex.One);
            KeyboardState kbState = Keyboard.GetState();
            // Handle inputs for the sample camera
            camera.HandleDefaultGamepadControls(gpState, gameTime);
            camera.HandleDefaultKeyboardControls(kbState, gameTime);
            // The built-in camera class provides the view matrix
            grid.ViewMatrix = camera.ViewMatrix;

            base.Update(gameTime);
        }

        #region UpdateInput
        protected void UpdateInput()
        {
            if (inputs.PauseGame)
            {
                this.Exit();
            }

            // Toggle animation?
            if (inputs.IsKeyPressed(Keys.A))
            {
                isAnimation = !isAnimation;
            }

            // Toggle wireframe?
            if (inputs.IsKeyPressed(Keys.W))
            {
                isWireframe = !isWireframe;
            }

            // Change primitive?
            if (inputs.IsKeyPressed(Keys.Space))
            {
                activePrimitive = (activePrimitive + 1) % primitives.Count;
            }
           
            // Change Y angle?
            if (inputs.IsKeyPressed(Keys.X))
            {
                angle += 5.0f;
            }

            // Change Translate?
            if (inputs.IsKeyPressed(Keys.Y))
            {
                zoom -= 0.1f;
            }
        }
        #endregion

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.Black);

            // draw the reference grid so it's easier to get our bearings
            grid.Draw();

            if (isWireframe)
            {
                GraphicsDevice.RasterizerState = wireFrameState;
            }
            else
            {
                GraphicsDevice.RasterizerState = RasterizerState.CullCounterClockwise;
            }

            // world matrix
            if (isAnimation) angle += 0.8f;
            world = Matrix.CreateFromYawPitchRoll(MathHelper.ToRadians(angle), 45.0f, 0.0f) * Matrix.CreateTranslation(0, 0, zoom);

            // Draw the current primitive.
            GeometricPrimitive currentPrimitive = primitives[activePrimitive];

            effect.World = world;
            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Apply();
                currentPrimitive.Draw(effect);
            }

            base.Draw(gameTime);
        }
    }
}
